/**
 * 2008, Digitalis Informatica. All rights reserved. Distribuicao e Gestao de Informatica, Lda. Estrada de Paco de Arcos
 * num.9 - Piso -1 2780-666 Paco de Arcos Telefone: (351) 21 4408990 Fax: (351) 21 4408999 http://www.digitalis.pt
 */

package pt.digitalis.utils.common;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.commons.beanutils.NestedNullException;
import org.apache.commons.beanutils.PropertyUtils;

/**
 * Inspects a given bean and it's properties
 * 
 * @author Pedro Viegas <a href="mailto:pviegas@digitalis.pt">pviegas@digitalis.pt</a><br/>
 * @created 2008/11/28
 */
public class BeanInspector {

    /**
     * Inspects and returns the class of the attribute of the given class
     * 
     * @param parentClass
     *            class that owns the attribute to inspect
     * @param attributePath
     *            the path of the attribute
     * @return the class of the desired attribute
     * @throws NoSuchFieldException
     * @throws SecurityException
     * @throws ClassNotFoundException
     */
    public static Class<?> getAttributeClass(Class<?> parentClass, String attributePath) throws SecurityException,
            NoSuchFieldException, ClassNotFoundException
    {
        Field field = null;

        while (attributePath.contains("."))
        {
            String[] parts = attributePath.split("\\.", 2);
            field = parentClass.getDeclaredField(parts[0]);
            parentClass = getAttributeClassIgnoringSetsOrLists(field);

            attributePath = parts[1];
        }

        field = parentClass.getDeclaredField(attributePath);

        return getAttributeClassIgnoringSetsOrLists(field);
    }

    /**
     * Will analyze the given field type and if it is a set or list will return the inner collection object type
     * 
     * @param field
     *            the field to validate
     * @return the filed type
     * @throws ClassNotFoundException
     */
    private static Class<?> getAttributeClassIgnoringSetsOrLists(Field field) throws ClassNotFoundException
    {
        Class<?> clazz = field.getType();

        if (Set.class.getCanonicalName().equals(clazz.getName())
                || List.class.getCanonicalName().equals(clazz.getName()))
        {
            String genericString = field.toGenericString();

            if (genericString.contains("<"))
            {
                String className = genericString.substring(genericString.indexOf("<") + 1, genericString.indexOf(">"));
                clazz = Class.forName(className);
            }
            else
                clazz = Object.class;
        }

        return clazz;
    }

    /**
     * Inspects and returns an instance of the class of the attribute of the given class
     * 
     * @param parentClass
     *            class that owns the attribute to inspect
     * @param attributePath
     *            the path of the attribute
     * @return an instance of the class of the desired attribute
     * @throws NoSuchFieldException
     * @throws SecurityException
     * @throws IllegalAccessException
     * @throws InstantiationException
     * @throws ClassNotFoundException
     */
    public static Object getAttributeClassInstance(Class<?> parentClass, String attributePath)
            throws SecurityException, NoSuchFieldException, InstantiationException, IllegalAccessException,
            ClassNotFoundException
    {
        Class<?> clazz = getAttributeClass(parentClass, attributePath);

        return clazz.newInstance();
    }

    /**
     * Inspects a given object and returns a map of properties and they values
     * 
     * @param obj
     *            the object to inspect
     * @return the map of attributes
     */
    @SuppressWarnings("unchecked")
    public static Map<String, String> getAttributeValues(Object obj)
    {

        Map<String, String> results = new HashMap<String, String>();

        if (obj instanceof String)
        {
            results.put("value", obj.toString());

        }
        else
        {
            Map<String, Object> props;
            try
            {
                props = PropertyUtils.describe(obj);

            }
            catch (IllegalAccessException e)
            {
                return results;
            }
            catch (InvocationTargetException e)
            {
                return results;
            }
            catch (NoSuchMethodException e)
            {
                return results;
            }

            for (Entry<String, Object> entry: props.entrySet())
            {
                if (!entry.getKey().equals("class") && !entry.getKey().equals("bytes")
                        && !(entry.getValue() instanceof Map) && !(entry.getValue() instanceof List)
                        && !(entry.getValue() instanceof Set))
                {
                    if (entry.getValue() == null)
                        results.put(entry.getKey(), "");
                    else
                        results.put(entry.getKey(), entry.getValue().toString());
                }
            }
        }

        return results;
    }

    /**
     * Retrieves the attribute from the object, even if is is a non-bean and/or private value
     * 
     * @param obj
     *            the object to retrieve the value from
     * @param propertyPath
     *            the path to the attribute to access
     * @return the property value
     */
    public static Object getPrivateAttributeValue(Object obj, String propertyPath)
    {
        if (obj == null || propertyPath == null || "".equals(propertyPath))
            return null;
        else
        {
            String[] parts = propertyPath.split("\\.");

            try
            {
                Field field;

                for (int i = 0; i < parts.length; i++)
                {
                    field = obj.getClass().getDeclaredField(parts[i]);

                    if (!field.isAccessible())
                        field.setAccessible(true);

                    obj = field.get(obj);
                }

                return obj;
            }
            catch (SecurityException e)
            {
                e.printStackTrace();
                return null;
            }
            catch (NoSuchFieldException e)
            {
                e.printStackTrace();
                return null;
            }
            catch (IllegalArgumentException e)
            {
                e.printStackTrace();
                return null;
            }
            catch (IllegalAccessException e)
            {
                e.printStackTrace();
                return null;
            }
        }
    }

    /**
     * Retrieves the attribute from the object, even if is is a non-bean and/or private value
     * 
     * @param obj
     *            the object to retrieve the value from
     * @param propertyPath
     *            the path to the attribute to access
     * @return the property value converted to a string
     */
    public static String getPrivateAttributeValueAsString(Object obj, String propertyPath)
    {
        return StringUtils.toStringOrNull(getPrivateAttributeValue(obj, propertyPath));
    }

    /**
     * Retrieves the attribute from the object
     * 
     * @param obj
     *            the object to retrieve the value from
     * @param propertyPath
     *            the path to the attribute to access
     * @return the property value
     */
    public static Object getValue(Object obj, String propertyPath)
    {

        if (obj == null)
            return null;
        else
        {
            try
            {
                String[] parts = propertyPath.split("toArray\\[", 2);

                // References collection
                if (parts.length > 1)
                {
                    if ("".equals(parts[0]))
                    {
                        // toArray is the start. No previous prop to read...
                        parts = parts[1].split("\\]", 2);

                        int index = Integer.parseInt(parts[0]);
                        String property = parts[1];

                        Collection<?> objCollection = (Collection<?>) obj;
                        Object objInCollection = null;

                        if (objCollection != null)
                        {
                            Object[] objArray = objCollection.toArray();

                            if (objArray.length > index)
                                objInCollection = objArray[index];
                        }

                        if (objInCollection == null)
                            return null;

                        if ("".equals(property))
                            // No more to parse. Array access only.
                            return objInCollection;

                        else
                        {
                            if (property.startsWith("."))
                                property = property.substring(1);

                            return getValue(objInCollection, property);
                        }
                    }
                    else
                    {
                        // Has previous path before toArray...
                        String before = parts[0];
                        String after = "toArray[" + parts[1];

                        if (before.endsWith("."))
                            before = before.substring(0, before.length() - 1);

                        obj = getValue(obj, before);

                        if (obj != null)
                            obj = getValue(obj, after);

                        return obj;
                    }
                }
                else if (propertyPath.contains("."))
                    // Nested property
                    return PropertyUtils.getNestedProperty(obj, propertyPath);
                else
                    // Direct attribute
                    return PropertyUtils.getProperty(obj, propertyPath);

            }
            catch (IllegalAccessException e)
            {
                return "";
            }
            catch (InvocationTargetException e)
            {
                return "";
            }
            catch (NoSuchMethodException e)
            {
                return "";
            }
            catch (NestedNullException e)
            {
                return "";
            }
        }
    }

    /**
     * Retrieves the attribute from the object
     * 
     * @param obj
     *            the object to retrieve the value from
     * @param propertyPath
     *            the path to the attribute to access
     * @return the property value
     */
    public static String getValueAsString(Object obj, String propertyPath)
    {

        Object value = getValue(obj, propertyPath);

        if (value == null)
            return null;
        else
            return value.toString();
    }

    /**
     * Instantiates the attribute object path given by the attributeName in the obj instance. Ex: call with
     * "innerClassAttr.secondLevelInnerClassAttr.attributeName" will instantiate first two classes:
     * <ul>
     * <li>obj.innerClassAttr = innerclassAttr.class.newInstance();</li>
     * <li>obj.innerClassAttr.secondLevelInnerClassAttr = secondLevelInnerClassAttr.class.newInstance();</li>
     * <li>return obj;</li>
     * <ul>
     * 
     * @param attributeName
     *            the attribute path to initialize (the last term is the attribute name)
     * @param obj
     *            the object instance
     * @return the obj instance fully initialized
     */
    public static Object instantiateAttributePath(String attributeName, Object obj)
    {
        Object innerAttribute = obj;
        String remainingPath = attributeName;

        try
        {
            int separatorPosition;

            // Crawl the attribute path and create the inner objects
            do
            {
                separatorPosition = remainingPath.indexOf(".");
                String currentAttributeName;
                Object value;

                if (separatorPosition != -1)
                {
                    currentAttributeName = remainingPath.substring(0, separatorPosition);
                    remainingPath = remainingPath.substring(separatorPosition + 1);
                }
                else
                {
                    currentAttributeName = remainingPath;
                }

                value = getAttributeClassInstance(innerAttribute.getClass(), currentAttributeName);
                Method method = innerAttribute.getClass().getMethod(
                        "set" + StringUtils.toUpperFirstChar(currentAttributeName), value.getClass());
                method.invoke(innerAttribute, value);

                innerAttribute = value;
            }
            while (separatorPosition != -1);

        }
        catch (Exception e)
        {
            e.printStackTrace();
            obj = null;
        }

        return obj;
    }

    /**
     * Sets a nested attribute in the given object
     * 
     * @param object
     *            the object to set the attribute on
     * @param attributeName
     *            the name of the nested attribute
     * @param value
     *            the value to set
     * @return the object instance with the attribute set
     */
    public static Object setNestedAtributeValue(Object object, String attributeName, Object value)
    {
        Object innerAttribute = object;
        String remainingPath = attributeName.replace("_", ".");

        try
        {
            // If it is an inner attribute, crawl the attribute path and create the inner objects
            while (remainingPath.contains("."))
            {
                int separatorPosition = remainingPath.indexOf(".");
                String currentAttributeName = remainingPath.substring(0, separatorPosition);

                Method methodGet = innerAttribute.getClass().getMethod(
                        "get" + StringUtils.toUpperFirstChar(currentAttributeName));
                Object currentAttributeObj = methodGet.invoke(innerAttribute);

                if (currentAttributeObj == null)
                {
                    currentAttributeObj = getAttributeClassInstance(innerAttribute.getClass(), currentAttributeName);
                }

                // Set the current object
                Method method = innerAttribute.getClass().getMethod(
                        "set" + StringUtils.toUpperFirstChar(currentAttributeName), currentAttributeObj.getClass());
                method.invoke(innerAttribute, currentAttributeObj);

                remainingPath = remainingPath.substring(separatorPosition + 1);
                innerAttribute = currentAttributeObj;
            }

            // Set the value of the single or last attribute
            Method method = innerAttribute.getClass().getMethod("set" + StringUtils.toUpperFirstChar(remainingPath),
                    value.getClass());
            method.invoke(innerAttribute, value);

        }
        catch (Exception e)
        {
            e.printStackTrace();
            object = null;
        }

        return object;
    }
}
