/**
 *
 */
package pt.digitalis.utils.common;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Locale;

/**
 * Utility for date object manipulation
 * 
 * @author Pedro Viegas <a href="mailto:pviegas@digitalis.pt">pviegas@digitalis.pt</a>
 * @created Nov 28, 2007
 */
public class DateUtils {

    /**
     * The date format to be used for conversions
     */
    private static String DATE_FORMAT = "EEE, d MMM yyyy HH:mm:ss.SS";

    /** Day unit */
    public static final String DAY = "day";

    /**
     * The date format to be used for conversions of javaScript date representations
     */
    private static String JAVASCRIPT_DATE_FORMAT = "yyyy-MM-dd HH:mm:ss";

    /** Month unit */
    public static final String MONTH = "month";

    /**
     * The date format to be used for conversions of simple dates
     */
    private static String SIMPLE_DATE_FORMAT = "dd/MM/yyyy";

    /**
     * The date format to be used for conversions of simple dates relative
     */
    private static String SIMPLE_DATE_FORMAT_RELATIVE = "dd/MM/yy";

    /**
     * The date format (with time) to be used for conversions of simple dates
     */
    private static String SIMPLE_DATE_FORMAT_TIME = "dd/MM/yyyy HH:mm:ss";

    /** Week unit */
    public static final String WEEK = "week";

    /** Year unit */
    public static final String YEAR = "year";

    /**
     * Converts a date to a string representation on a conventioned format
     * 
     * @param date
     *            the date to convert
     * @return the date as a string
     */
    static public String dateToString(Date date)
    {
        DateFormat dateFormat = new SimpleDateFormat(DATE_FORMAT);
        return dateFormat.format(date);
    }

    /**
     * Get diff in days between two dates
     * 
     * @param dateInicial
     *            the initial date
     * @param dateFinal
     *            the final date
     * @return days
     */
    public static Long getDateDiffInDays(Date dateInicial, Date dateFinal)
    {
        return (dateFinal.getTime() - dateInicial.getTime()) / 86400000;
    }

    /**
     * Return the difference between two dates in months
     * 
     * @param dateInicial
     * @param dateFinal
     * @return the month difference
     */
    static public Integer getDateDiffInMonth(Date dateInicial, Date dateFinal)
    {
        Integer diffMonths = 0;
        Calendar temp = Calendar.getInstance();
        temp.setTime(dateInicial);
        while (temp.getTime().before(dateFinal))
        {
            diffMonths++;
            temp.add(Calendar.MONTH, 1);
        }
        return diffMonths;
    }

    /**
     * Return the difference between two dates in years
     * 
     * @param date
     * @return the years difference
     */
    static public Integer getDateDiffInYears(Date date)
    {
        Calendar temp = Calendar.getInstance();
        Date comparisonDate = temp.getTime();
        boolean invert = date.before(temp.getTime());
        Integer diffYears = 0;

        // Date is in the past, must return negative number
        if (invert)
            temp.setTime(date);
        else
            comparisonDate = date;

        while (temp.getTime().before(comparisonDate))
        {
            diffYears++;
            temp.add(Calendar.YEAR, 1);
        }

        diffYears--;

        if (invert)
            diffYears = -diffYears;

        return diffYears;
    }

    /**
     * @return the date/time format used for conversions by this lib
     */
    static public DateFormat getDateTimeFormat()
    {
        return new SimpleDateFormat(SIMPLE_DATE_FORMAT_TIME);
    }

    /**
     * @return the long date format used for conversions by this lib
     */
    static public DateFormat getLongFormat()
    {
        return new SimpleDateFormat(DATE_FORMAT);
    }

    /**
     * Return the Month Name for a given date
     * 
     * @param date
     * @return the Month Name
     */
    static public String getMonthName(Date date)
    {
        return getMonthName(date, Locale.getDefault().getLanguage());
    }

    /**
     * Return the Month Name for a given date in the given language
     * 
     * @param date
     * @param language
     *            The given language
     * @return the Month Name
     */
    static public String getMonthName(Date date, String language)
    {
        Locale locale = new Locale(language);
        DateFormat df = new SimpleDateFormat("MMMMM", locale);
        return df.format(date);
    }

    /**
     * Return the Month Name for a given Integer Month
     * 
     * @param monthInt
     *            The Month Number
     * @return the Month Name
     */
    static public String getMonthName(Integer monthInt)
    {
        Calendar cal = Calendar.getInstance();
        cal.set(Calendar.DAY_OF_MONTH, 1);
        cal.set(Calendar.MONTH, monthInt - 1);
        return getMonthName(cal.getTime(), Locale.getDefault().getLanguage());
    }

    /**
     * Return the Month Name for a given Integer Month in the given language
     * 
     * @param monthInt
     *            The Month Number
     * @param language
     *            The given language
     * @return the Month Name
     */
    static public String getMonthName(Integer monthInt, String language)
    {
        Calendar cal = Calendar.getInstance();
        cal.set(Calendar.MONTH, monthInt - 1);
        return getMonthName(cal.getTime(), language);
    }

    /**
     * @return the simple date format used for conversions by this lib
     */
    static public DateFormat getShortFormat()
    {
        return new SimpleDateFormat(SIMPLE_DATE_FORMAT);
    }

    /**
     * @return a date object without time information. trimmed to 0:00:00
     */
    static public Date getTrimmedDate()
    {
        try
        {
            return stringToSimpleDate(simpleDateToString(new Date()));
        }
        catch (ParseException e)
        {
            return null;
        }
    }

    /**
     * Calculates the max value for a given date in the given unit
     * 
     * @param date
     *            the date to maximize
     * @param unit
     *            "week", "month" or "year"
     * @return the maximized date
     */
    static public Date maxDate(Date date, String unit)
    {

        return truncDate(date, unit);
    }

    /**
     * Calculates the min value for a given date in the given unit
     * 
     * @param date
     *            the date to minimize
     * @param unit
     *            "week", "month" or "year"
     * @return the minimized date
     */
    static public Date minDate(Date date, String unit)
    {
        Calendar temp = Calendar.getInstance();
        temp.setTime(date);

        if (unit.equals(WEEK))
        {
            temp.set(Calendar.DAY_OF_WEEK, Calendar.getInstance().getFirstDayOfWeek());
            return temp.getTime();

        }
        else if (unit.equals(MONTH))
        {
            temp.set(Calendar.DAY_OF_MONTH, 1);
            return temp.getTime();

        }
        else if (unit.equals(YEAR))
        {
            temp.set(Calendar.DAY_OF_YEAR, 1);
            return temp.getTime();

        }
        else
            return date;
    }

    /**
     * verts a date to a string representation on a conventioned format with time
     * 
     * @param date
     *            the date to convert
     * @return the date as a string
     */
    static public String simpleDateTimeToString(Date date)
    {
        DateFormat simpleDateFormat = new SimpleDateFormat(SIMPLE_DATE_FORMAT_TIME);
        return simpleDateFormat.format(date);
    }

    /**
     * Con* Converts a date to a string representation on a conventioned format
     * 
     * @param date
     *            the date to convert
     * @return the date as a string
     */
    static public String simpleDateToString(Date date)
    {
        DateFormat simpleDateFormat = new SimpleDateFormat(SIMPLE_DATE_FORMAT);
        return simpleDateFormat.format(date);
    }

    /**
     * * Converts a date to a string representation on a conventioned format
     * 
     * @param date
     *            the date string representation to convert
     * @return the date
     * @throws ParseException
     *             if the string does not respect the conventioned date pattern
     */
    static public Date stringToDate(String date) throws ParseException
    {
        if (date == null || "".equals(date.trim()))
            return null;
        else
        {
            DateFormat dateFormat;
            if (date.matches("^[0-9]{4}-(0[0-9]|1[0-2])-([012][0-9]|3[01])T([0-2][0-3]:[0-5][0-9]:[0-5][0-9])$"))
            {
                date = date.replace("T", " ");
                dateFormat = new SimpleDateFormat(JAVASCRIPT_DATE_FORMAT);
            }

            else
                dateFormat = new SimpleDateFormat(DATE_FORMAT);

            return dateFormat.parse(date.trim());
        }

    }

    /**
     * Converts a date to a string representation on a conventioned simple format. Just dd/MM/yy
     * 
     * @param date
     *            the date string representation to convert
     * @return the date
     * @throws ParseException
     *             if the string does not respect the conventioned date pattern
     */
    static public Date stringToSimpleDate(String date) throws ParseException
    {
        if (date != null)
        {
            if (date.length() > 10)
                // not a simple format. revert to general function
                return stringToDate(date);
            else
            {
                // Simple format
                DateFormat simpleDateFormat = null;
                if (date.length() > 8)
                    simpleDateFormat = new SimpleDateFormat(SIMPLE_DATE_FORMAT);
                else if (date.length() > 10)
                    return stringToDate(date);
                else
                    simpleDateFormat = new SimpleDateFormat(SIMPLE_DATE_FORMAT_RELATIVE);
                return simpleDateFormat.parse(date);
            }
        }
        else
        {
            return null;
        }
    }

    /**
     * Converts a date to a string representation on a conventioned simple format. Just dd/MM/yy In case the given date
     * is null or invalid in format will return null without any exception raised.
     * 
     * @param date
     *            the date string representation to convert
     * @return the date
     */
    static public Date stringToSimpleDateOrNull(Object date)
    {
        Date result = null;

        if (date != null)
        {
            try
            {
                result = stringToSimpleDate(date.toString());
            }
            catch (Exception e)
            {
            }
        }

        return result;
    }

    /**
     * Truncates a date to the given unit
     * 
     * @param date
     *            the date to truncate
     * @param unit
     *            "week", "month" or "year"
     * @return the truncated date
     */
    static public Date truncDate(Date date, String unit)
    {
        Calendar temp = Calendar.getInstance();
        temp.setTime(date);

        if (unit.equals(WEEK))
        {

            if (Calendar.getInstance().getFirstDayOfWeek() == Calendar.MONDAY)
                temp.set(Calendar.DAY_OF_WEEK, Calendar.SUNDAY);
            else
                temp.set(Calendar.DAY_OF_WEEK, Calendar.SATURDAY);

            return temp.getTime();

        }
        else if (unit.equals(MONTH))
        {
            temp.set(Calendar.DAY_OF_MONTH, temp.getActualMaximum(Calendar.DAY_OF_MONTH));
            return temp.getTime();

        }
        else if (unit.equals(YEAR))
        {
            temp.set(Calendar.MONTH, Calendar.DECEMBER);
            temp.set(Calendar.DAY_OF_MONTH, 31);
            return temp.getTime();

        }
        else
            return date;
    }
}
