/**
 * 2007, Digitalis Informatica. All rights reserved. Distribuicao e Gestao de Informatica, Lda. Estrada de Paco de Arcos
 * num.9 - Piso -1 2780-666 Paco de Arcos Telefone: (351) 21 4408990 Fax: (351) 21 4408999 http://www.digitalis.pt
 */
package pt.digitalis.utils.common.collections;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;

/**
 * This class implements a cross-over ADT that mixes the FIFO discipline access with the ability of searching through
 * the elements without iterating over them, like it's typical for FIFO access. This strategy aims for fast lookup and
 * modifying operations while keeping the queue's natural access discipline. <br>
 * The ADT is generic, so it can work with any kind of objects as keys or values. <br>
 * It is thread-safe due to the use of thread-safe containers.
 *
 * @author Rodrigo Gonalves <a href="mailto:rgoncalves@digitalis.pt">rgoncalves@digitalis.pt</a><br/>
 * @param <Key>
 *            the type of the key
 * @param <Value>
 *            the type of the value
 */
public class SearchableQueue<Key, Value extends Object> {

    /** The element that guarantees the FIFO access. It's thread-safe. */
    Queue<Key> order = new ConcurrentLinkedQueue<Key>();

    /**
     * The element that guarantees fast search capabilities. This attribute is also the main element storage. It's
     * thread-safe.
     */
    Map<Key, Entry> storage = new HashMap<Key, Entry>();

    /**
     * Holds the absolute position of the last element in queue.
     */
    long fifoQueueLastElementAbsolutePosition = 0;

    /**
     * Inserts a new object into this <code>SearchableQueue</code>
     *
     * @param key
     *            the element key
     * @param value
     *            the element value
     * @return T if it was possible to add the element, F otherwise
     */
    public boolean offer(Key key, Value value)
    {

        if (storage.containsKey(key))

            // When the key already exists in the Searchable Queue then it cnnot
            // be inserted. This is due to the fact that a Queued element cannot
            // be updated
            return false;
        else
        {
            // Move the first element one position forward
            ++fifoQueueLastElementAbsolutePosition;

            // Create a new entry
            Entry entry = new Entry(value, fifoQueueLastElementAbsolutePosition);

            // Store the entry on the <code>storage</code> container
            storage.put(key, entry);
            // Store the key on the FIFO queue (last position!!).
            order.offer(key);

            return true;
        }
    }

    /**
     * Retrieves and removes the head of this queue.
     *
     * @return the element on the head of the queue
     * @throws NoSuchElementException
     *             if the FIFO Queue is empty
     */
    @SuppressWarnings("unchecked")
    public Value remove() throws NoSuchElementException
    {
        // Return the value indexed by the key at the head of this queue,
        // removing the values from the storage and it's key from the queue as
        // well.
        return (Value) (storage.remove(order.remove()).getValue());
    }

    /**
     * Retrieves but doesn't remove the head of this queue.
     *
     * @return the element on the head of the queue
     * @throws NoSuchElementException
     *             if the FIFO Queue is empty
     */
    @SuppressWarnings("unchecked")
    public Value element() throws NoSuchElementException
    {
        return (Value) (storage.get(order.element())).getValue();
    }

    /**
     * Retrieves, but doesn't remove, the head of this queue, returning null if this queue is empty.
     *
     * @return the head of this queue, or null if this queue is empty.
     */
    @SuppressWarnings("unchecked")
    public Value peek()
    {
        return (Value) (storage.get(order.peek()).getValue());
    }

    /**
     * Retrieves and removes the head of this queue, or null if this queue is empty.
     *
     * @return the head of this queue
     */
    @SuppressWarnings("unchecked")
    public Value poll()
    {
        return (Value) (storage.remove(order.poll()).getValue());
    }

    /**
     * Returns true if the supplied key exists on the <code>searchableQueue</code>.
     *
     * @param key
     *            the key to search for
     * @return T if the <code>searchableQueue</code> contains the given key, F otherwise.
     */
    public boolean containsKey(Key key)
    {
        return storage.containsKey(key);
    }

    /**
     * Returns the position on the FIFO queue of the element with the supplied key. The position is obtained by
     * subtracting the current sentinel value added to the position at insertion from the SearchableQueue size. <br>
     * Returns -1 if the <code>SearchableQueue</code> doesn't contain the element.
     *
     * @param key
     *            the element's key
     * @return the position of the element with the given id on the FIFO queue if exists, -1 otherwise
     */
    public long getPosition(Key key)
    {
        Entry entry = storage.get(key);

        if (entry != null)
            return (storage.size() - (fifoQueueLastElementAbsolutePosition - entry.getPosition()));
        else
            return -1;
    }

    /**
     * Returns the value with the given key if it exists on the <code>SearchableQueue</code>.
     *
     * @param key
     *            the value's key
     * @return the value of the given key
     */
    @SuppressWarnings("unchecked")
    public Value getValue(Key key)
    {
        return (Value) (storage.get(key)).getValue();
    }

    /**
     * Returns the size of the <code>searchableQueue</code>.
     *
     * @return the size of the <code>searchableQueue</code>
     */
    public int size()
    {
        return order.size();
    }

    /**
     * Returns true if the <code>searchableQueue</code> is empty.
     *
     * @return T if the the <code>searchableQueue</code> is empty, F otherwise
     */
    public boolean isEmpty()
    {
        return order.isEmpty();
    }

    /**
     * Provides a string representation for the <code>SearchableQueue</code>. The format is: {First element position
     * value is: <fifoQueueLastElementAbsolutePosition>} (position=<position>, key=<key>, value=<value>) ...
     *
     * @return the SearchableQueue string representation.
     */
    public String toString()
    {

        // Initialize the result String with management information
        String result = "\n[SearchableQueue contents]\n";
        result += "  |--  Elements in queue: " + this.size() + "\n";
        result += "  |--  Last element absolute position: " + fifoQueueLastElementAbsolutePosition + "\n";

        if (this.isEmpty())
            result += "  `- [No elements in queue]\n";
        else
        {
            result += "  `- [Element dump]\n";

            // A temporary String to hold the triplet currently pointed by the
            // iterator
            String temp;

            // Prevents changes during iteration
            synchronized (storage)
            {
                Iterator<Key> iter = order.iterator();
                int i = 0;

                // Iterate...
                while (iter.hasNext())
                {
                    i++;
                    Key key = iter.next();
                    Entry entry = storage.get(key);

                    // Compose output string...
                    if (i < this.size())
                        temp = "         |--  ";
                    else
                        temp = "         `--  ";

                    temp += "(position=" + getPosition(key) + ", key=" + key + ", value=" + entry.getValue() + ")\n";

                    result += temp;
                }
            }
        }

        // Return result
        return result;
    }

    /**
     * This class represents an Entry of the <code>SearchableQueue</code>. It stores the object and it's position on
     * the FIFO queue.
     */
    static private class Entry {

        /** The entry value. */
        Object value;

        /** The entry position on the FIFO queue. */
        long position;

        /**
         * Constructor. Builds a new entry.
         *
         * @param element
         *            the element to store
         * @param position
         *            the element's position on the FIFO queue
         */
        public Entry(Object element, long position)
        {
            this.value = element;
            this.position = position;
        }

        /**
         * Returns the value of this entry
         *
         * @return the value of this entry
         */
        final public Object getValue()
        {
            return this.value;
        }

        /**
         * Returns the entry position on the FIFO queue.
         *
         * @return the entry position on the FIFO queue
         */
        final public long getPosition()
        {
            return this.position;
        }
    }
}
